from telegrinder.bot.cute_types.message import MessageCute
from telegrinder.node.base import ComposeError, scalar_node


@scalar_node
class ReplyMessage:
    @classmethod
    def compose(cls, message: MessageCute) -> MessageCute:
        return message.reply_to_message.expect(ComposeError("Message doesn't have reply"))


__all__ = ("ReplyMessage",)



import sys

from telegrinder import Dispatch, Message
from telegrinder.rules import ABCRule, Text


class IsAdmin(ABCRule):
    def check(self, message: Message) -> bool:
        return message.from_user.username.unwrap_or_none() == "timoniq"


dp = Dispatch()
dp.message.auto_rules = IsAdmin()


@dp.message(Text("/explode"))
async def explode_handler(message: Message):
    await message.answer("Oops bot exploded =(")
    sys.exit()



import os
import typing

from choicelib import choice_in_order

from telegrinder.msgspec_utils import json

if typing.TYPE_CHECKING:
    from logging import Handler as LoggingBasicHandler

    from loguru import AsyncHandlerConfig as LoguruAsyncHandlerConfig  # type: ignore
    from loguru import BasicHandlerConfig as LoguruBasicHandlerConfig  # type: ignore
    from loguru import FileHandlerConfig as LoguruFileHandlerConfig  # type: ignore
    from loguru import HandlerConfig as LoguruHandlerConfig  # type: ignore
else:
    LoguruAsyncHandlerConfig = LoguruBasicHandlerConfig = LoguruFileHandlerConfig = dict

type _LoggerHandler = LoggingBasicHandler | LoguruHandlerConfig


def _remove_handlers(logger: typing.Any, /) -> None:
    for hdlr in logger.handlers[:]:
        logger.removeHandler(hdlr)


@typing.runtime_checkable
class LoggerModule(typing.Protocol):
    def debug(self, __msg: typing.Any, *args: typing.Any, **kwargs: typing.Any) -> None: ...

    def info(self, __msg: typing.Any, *args: typing.Any, **kwargs: typing.Any) -> None: ...

    def warning(self, __msg: typing.Any, *args: typing.Any, **kwargs: typing.Any) -> None: ...

    def error(self, __msg: typing.Any, *args: typing.Any, **kwargs: typing.Any) -> None: ...

    def critical(self, __msg: typing.Any, *args: typing.Any, **kwargs: typing.Any) -> None: ...

    def exception(self, __msg: typing.Any, *args: typing.Any, **kwargs: typing.Any) -> None: ...

    if typing.TYPE_CHECKING:

        def set_level(
            self,
            level: typing.Literal[
                "DEBUG",
                "INFO",
                "WARNING",
                "ERROR",
                "CRITICAL",
                "EXCEPTION",
            ],
            /,
        ) -> None: ...

        def set_new_handler(self, new_handler: _LoggerHandler, /) -> None: ...


logger: LoggerModule
logging_level = os.getenv("LOGGER_LEVEL", default="DEBUG").upper()
logging_module = choice_in_order(["structlog", "loguru"], default="logging", do_import=False)
asyncio_module = choice_in_order(["uvloop", "winloop"], default="asyncio", do_import=False)


if logging_module == "loguru":
    import os
    import sys

    from loguru import logger  # type: ignore

    os.environ.setdefault("LOGURU_AUTOINIT", "0")
    log_format = (
        "<level>{level: <8}</level> | "
        "<lg>{time:YYYY-MM-DD HH:mm:ss}</lg> | "
        "<le>{name}</le>:<le>{function}</le>:"
        "<le>{line}</le> > <lw>{message}</lw>"
    )
    logger.remove()  # type: ignore
    handler_id = logger.add(  # type: ignore
        sink=sys.stderr,
        format=log_format,
        enqueue=True,
        colorize=True,
        level=logging_level,
    )

elif logging_module == "structlog":
    import logging
    import re
    import sys
    import typing
    from contextlib import suppress

    import colorama
    import structlog  # type: ignore

    LEVELS_COLORS = dict(
        debug=colorama.Fore.LIGHTBLUE_EX,
        info=colorama.Fore.LIGHTGREEN_EX,
        warning=colorama.Fore.LIGHTYELLOW_EX,
        error=colorama.Fore.LIGHTRED_EX,
        critical=colorama.Fore.LIGHTRED_EX,
    )

    class SLF4JStyleFormatter:
        TOKENS = frozenset(("{}", "{!s}", "{!r}", "{!a}"))
        TOKENS_PATTERN = re.compile(r"({!r}|{})")

        def __init__(self, *, remove_positional_args: bool = True) -> None:
            self.remove_positional_args = remove_positional_args

        def __call__(
            self,
            logger: typing.Any,
            method_name: str,
            event_dict: dict[str, typing.Any],
        ) -> dict[str, typing.Any]:
            args = event_dict.get("positional_args")
            if not args:
                return event_dict

            event = event_dict.get("event", "")
            if not isinstance(event, str):
                return event_dict

            log_level = event_dict.get("level", "debug")
            with suppress(TypeError, ValueError, IndexError):
                if "{}" in event or "{!r}" in event:
                    event_dict["event"] = self._safe_format_braces(event, args, log_level)
                elif len(args) == 1 and isinstance(args[0], dict):
                    formatted = event % args[0]
                    event_dict["event"] = self._highlight_values(
                        formatted, args[0].values(), log_level, percent_style=True
                    )
                else:
                    formatted = event % args
                    event_dict["event"] = self._highlight_values(formatted, args, log_level, percent_style=True)

            if self.remove_positional_args and "positional_args" in event_dict:
                del event_dict["positional_args"]

            return event_dict

        def _colorize(self, value: typing.Any, log_level: str) -> str:
            return f"{LEVELS_COLORS[log_level]}{value}{colorama.Fore.RESET}"

        def _safe_format_braces(
            self,
            message: str,
            args: tuple[typing.Any, ...],
            log_level: str,
        ) -> str:
            tokens = self.TOKENS_PATTERN.split(message)
            result = []
            arg_index = 0

            for token in tokens:
                if token in self.TOKENS and arg_index < len(args):
                    result.append(
                        self._colorize(
                            str(args[arg_index]) if token != "{!r}" else repr(args[arg_index]),
                            log_level,
                        ),
                    )
                    arg_index += 1
                else:
                    result.append(token)

            return "".join(result)

        def _highlight_values(
            self,
            full_message: str,
            values: typing.Iterable[typing.Any],
            log_level: str,
            percent_style: bool = False,
        ) -> str:
            for v in values:
                with suppress(Exception):
                    raw = repr(v) if percent_style and ("%r" in full_message) else str(v)
                    pattern = re.compile(rf"(?<!%)\b{re.escape(raw)}\b")
                    full_message = pattern.sub(
                        lambda m: self._colorize(m.group(0), log_level), full_message, count=1
                    )

            return full_message

    class LogLevelColumnFormatter:
        def __call__(self, key: str, value: typing.Any) -> str:
            color = LEVELS_COLORS[value]
            return f"[{color}{value:^12}{colorama.Fore.RESET}]"

    class Filter(logging.Filter):
        def filter(self, record: logging.LogRecord) -> bool:
            level_color = LEVELS_COLORS[record.levelname.lower()]
            location = (
                f"{colorama.Fore.LIGHTCYAN_EX}{record.module}{colorama.Fore.RESET}:"
                f"{level_color}{record.funcName}{colorama.Fore.RESET}:"
                f"{colorama.Fore.LIGHTMAGENTA_EX}{record.lineno}{colorama.Fore.RESET} "
            )
            record.location = location
            return True

    def configure_logging() -> None:
        console_renderer = structlog.dev.ConsoleRenderer(colors=True)

        for column in console_renderer._columns:
            if column.key == "level":
                column.formatter = LogLevelColumnFormatter()
                break

        structlog.configure(  # type: ignore
            processors=[
                structlog.stdlib.filter_by_level,
                structlog.stdlib.add_log_level,
                SLF4JStyleFormatter(),  # type: ignore
                structlog.processors.StackInfoRenderer(),
                structlog.processors.format_exc_info,
                structlog.processors.UnicodeDecoder(),
                console_renderer,
            ],
            context_class=dict,
            logger_factory=structlog.stdlib.LoggerFactory(),
            wrapper_class=structlog.stdlib.BoundLogger,
            cache_logger_on_first_use=True,
        )
        fmt = (
            f"[{colorama.Fore.LIGHTBLUE_EX}{{name}}{colorama.Style.RESET_ALL}] "
            f"{colorama.Fore.LIGHTWHITE_EX}{{location}}{colorama.Style.RESET_ALL}"
            f"[{colorama.Fore.LIGHTBLACK_EX}{{asctime}}{colorama.Style.RESET_ALL}] "
            f"{colorama.Fore.LIGHTWHITE_EX}~{colorama.Style.RESET_ALL} {{message}}"
        )

        telegrinder_logger = logging.getLogger("telegrinder")
        telegrinder_logger.setLevel(logging_level)

        handler = logging.StreamHandler(sys.stderr)
        handler.addFilter(Filter())
        formatter = logging.Formatter(fmt, datefmt="%Y-%m-%d %H:%M:%S", style="{")
        handler.setFormatter(formatter)
        _remove_handlers(telegrinder_logger)
        telegrinder_logger.addHandler(handler)

    configure_logging()
    logger = structlog.get_logger("telegrinder")  # type: ignore

elif logging_module == "logging":
    """
    This is a workaround for lazy formatting with {} in logging.
    About:
    https://docs.python.org/3/howto/logging-cookbook.html#use-of-alternative-formatting-styles
    """

    import inspect
    import logging
    import sys

    import colorama

    colorama.just_fix_windows_console()
    colorama.init()

    LOG_FORMAT = (
        "<light_white>{name: <4} |</light_white> <level>{levelname: <8}</level>"
        " <light_white>|</light_white> <light_green>{asctime}</light_green> <light_white>"
        "|</light_white> <level_module>{module}</level_module><light_white>:</light_white>"
        "<func_name>{funcName}</func_name><light_white>:</light_white><lineno>{lineno}</lineno>"
        " <light_white>></light_white> <message>{message}</message>"
    )
    COLORS = dict(
        reset=colorama.Style.RESET_ALL,
        red=colorama.Fore.RED,
        green=colorama.Fore.GREEN,
        blue=colorama.Fore.BLUE,
        white=colorama.Fore.WHITE,
        yellow=colorama.Fore.YELLOW,
        magenta=colorama.Fore.MAGENTA,
        cyan=colorama.Fore.CYAN,
        light_red=colorama.Fore.LIGHTRED_EX,
        light_green=colorama.Fore.LIGHTGREEN_EX,
        light_blue=colorama.Fore.LIGHTBLUE_EX,
        light_white=colorama.Fore.LIGHTWHITE_EX,
        light_yellow=colorama.Fore.LIGHTYELLOW_EX,
        light_magenta=colorama.Fore.LIGHTMAGENTA_EX,
        light_cyan=colorama.Fore.LIGHTCYAN_EX,
    )
    LEVEL_FORMAT_SETTINGS = dict(
        DEBUG=dict(
            level="light_blue",
            level_module="blue",
            func_name="blue",
            lineno="light_yellow",
            message="light_blue",
        ),
        INFO=dict(
            level="cyan",
            level_module="light_cyan",
            func_name="light_cyan",
            lineno="light_yellow",
            message="light_green",
        ),
        WARNING=dict(
            level="light_yellow",
            level_module="light_magenta",
            func_name="light_magenta",
            lineno="light_blue",
            message="light_yellow",
        ),
        ERROR=dict(
            level="red",
            level_module="light_yellow",
            func_name="light_yellow",
            lineno="light_blue",
            message="light_red",
        ),
        CRITICAL=dict(
            level="magenta",
            level_module="light_red",
            func_name="light_red",
            lineno="light_yellow",
            message="light_magenta",
        ),
    )
    LOG_FORMAT = (
        LOG_FORMAT.replace("<light_white>", COLORS["light_white"])
        .replace("</light_white>", COLORS["reset"])
        .replace("<light_green>", COLORS["light_green"])
        .replace("</light_green>", COLORS["reset"])
    )
    LEVEL_FORMATS = dict[str, str]()

    for level, settings in LEVEL_FORMAT_SETTINGS.items():
        fmt = LOG_FORMAT

        for name, color in settings.items():
            fmt = fmt.replace(f"<{name}>", COLORS[color]).replace(f"</{name}>", COLORS["reset"])

        LEVEL_FORMATS[level] = fmt

    class TelegrinderLoggingFormatter(logging.Formatter):
        def format(self, record: logging.LogRecord) -> str:
            if not record.funcName or record.funcName == "<module>":
                record.funcName = "\b"

            frame = sys._getframe(1)
            while frame:
                if frame.f_code.co_filename == record.pathname and frame.f_lineno == record.lineno:
                    break

                frame = frame.f_back

            if frame is not None:
                record.module = frame.f_globals.get("__name__", "<module>")

            return logging.Formatter(
                fmt=LEVEL_FORMATS.get(record.levelname),
                datefmt="%Y-%m-%d %H:%M:%S",
                style="{",
            ).format(record)

    class LogMessage:
        def __init__(self, fmt: str, args: typing.Any, kwargs: typing.Any) -> None:
            self.fmt = fmt
            self.args = args
            self.kwargs = kwargs

        def __str__(self) -> str:
            return self.fmt.format(*self.args, **self.kwargs)

    class TelegrinderLoggingStyleAdapter(logging.LoggerAdapter):
        logger: logging.Logger

        def __init__(
            self,
            logger: logging.Logger,
            **extra: typing.Any,
        ) -> None:
            super().__init__(logger, extra=extra)
            self.log_arg_names = frozenset(inspect.getfullargspec(self.logger._log).args[1:])

        def log(self, level: int, msg: typing.Any, *args: typing.Any, **kwargs: typing.Any) -> None:
            if self.isEnabledFor(level):
                msg, args, kwargs = self.proc(msg, args, kwargs)
                self.logger._log(level, msg, args, **kwargs)

        def proc(
            self,
            msg: typing.Any,
            args: tuple[typing.Any, ...],
            kwargs: dict[str, typing.Any],
        ) -> tuple[typing.Any, tuple[typing.Any, ...], dict[str, typing.Any]]:
            kwargs.setdefault("stacklevel", 2)

            if isinstance(msg, str):
                msg = LogMessage(msg, args, kwargs)
                args = tuple()

            return msg, args, {name: kwargs[name] for name in self.log_arg_names if name in kwargs}

    handler = logging.StreamHandler(sys.stderr)
    handler.setFormatter(TelegrinderLoggingFormatter())
    logger = logging.getLogger("telegrinder")  # type: ignore
    _remove_handlers(logger)
    logger.setLevel(logging_level)  # type: ignore
    logger.addHandler(handler)  # type: ignore
    logger = TelegrinderLoggingStyleAdapter(logger)  # type: ignore


if asyncio_module == "uvloop":
    import asyncio

    import uvloop  # type: ignore

    asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())  # type: ignore

elif asyncio_module == "winloop":
    import asyncio

    import winloop  # type: ignore

    asyncio.set_event_loop_policy(winloop.EventLoopPolicy())  # type: ignore


def _set_logger_level(level: str, /) -> None:
    level = level.upper()
    if logging_module in ("logging", "structlog"):
        import logging

        logging.getLogger("telegrinder").setLevel(level)
    elif logging_module == "loguru":
        import loguru  # type: ignore

        if handler_id in loguru.logger._core.handlers:  # type: ignore
            loguru.logger._core.handlers[handler_id]._levelno = loguru.logger.level(level).no  # type: ignore


def _set_logger_handler(new_handler: typing.Any, /) -> None:
    if logging_module in ("logging", "structlog"):
        import logging

        telegrinder_logger = logging.getLogger("telegrinder")
        _remove_handlers(telegrinder_logger)
        telegrinder_logger.addHandler(new_handler)
    elif logging_module == "loguru":
        import loguru  # type: ignore

        global handler_id  # type: ignore
        loguru.logger.remove(handler_id)  # type: ignore
        handler_id = loguru.logger.configure(handlers=(new_handler,))[0]  # type: ignore


setattr(logger, "set_level", staticmethod(_set_logger_level))
setattr(logger, "set_new_handler", staticmethod(_set_logger_handler))


__all__ = (
    "LoggerModule",
    "LoguruAsyncHandlerConfig",
    "LoguruBasicHandlerConfig",
    "LoguruFileHandlerConfig",
    "json",
    "logger",
)



import dataclasses
import pathlib
import typing

import msgspec
from tomlkit.toml_document import TOMLDocument
from tomlkit.toml_file import TOMLFile

type Path = str | pathlib.Path


@dataclasses.dataclass
class ConfigTOML:
    file: TOMLFile
    document: TOMLDocument

    def save(self) -> None:
        self.file.write(data=self.document)

    def as_model[T: msgspec.Struct](
        self,
        model_type: type[T],
        /,
        *,
        dec_hook: typing.Callable[[type[typing.Any], typing.Any], typing.Any] | None = None,
    ) -> T:
        return msgspec.convert(self.document.unwrap(), type=model_type, dec_hook=dec_hook)


def read_config(path: Path, /) -> ConfigTOML:
    file = TOMLFile(path)
    return ConfigTOML(file, document=file.read())


__all__ = ("ConfigTOML", "read_config")



from __future__ import annotations

import abc
import typing

from telegrinder.tools.keyboard.abc import ABCKeyboard, ABCStaticKeyboard, AnyMarkup, DictStrAny
from telegrinder.tools.keyboard.buttons.base import BaseButton

if typing.TYPE_CHECKING:
    from telegrinder.tools.keyboard.buttons.base import BaseStaticButton

KEYBOARD_BUTTONS_KEY: typing.Final[str] = "_buttons"
NO_VALUE: typing.Final[object] = object()


def _is_implemented[T: ABCKeyboard](
    value: object,
    kb_class: type[T],
    /,
) -> typing.TypeGuard[T | BaseButton]:
    return isinstance(value, BaseButton | kb_class)


class BaseKeyboard(ABCKeyboard, abc.ABC):
    @abc.abstractmethod
    def dict(self) -> DictStrAny:
        pass

    @abc.abstractmethod
    def get_markup(self) -> AnyMarkup:
        pass

    @abc.abstractmethod
    def copy(self, **with_changes: typing.Any) -> typing.Self:
        pass

    def __and__(self, other: object, /) -> typing.Self:
        if not _is_implemented(other, type(self)):
            return NotImplemented
        return self.add(other) if isinstance(other, BaseButton) else self.merge(other)

    def __or__(self, other: object, /) -> typing.Self:
        if not _is_implemented(other, type(self)):
            return NotImplemented

        kb = self.row()
        return kb.add(other) if isinstance(other, BaseButton) else kb.merge_to_last_row(other)


class BaseStaticKeyboard(ABCStaticKeyboard, abc.ABC):
    __max_in_row__: int

    @classmethod
    def get_buttons(cls) -> dict[str, BaseStaticButton]:
        if (buttons := cls._get_secret_value(KEYBOARD_BUTTONS_KEY)) is not NO_VALUE:
            return buttons
        return cls._set_secret_value(KEYBOARD_BUTTONS_KEY, super().get_buttons())

    @classmethod
    def _set_secret_value[V](cls, key: str, value: V, /) -> V:
        setattr(cls, key, value)
        return value

    @classmethod
    def _get_secret_value(cls, key: str, /) -> typing.Any:
        return getattr(cls, key, NO_VALUE)

    @abc.abstractmethod
    def dict(self) -> DictStrAny:
        pass

    @abc.abstractmethod
    def get_markup(self) -> AnyMarkup:
        pass


__all__ = ("BaseKeyboard", "BaseStaticKeyboard")



from telegrinder import API, Message, Telegrinder, Token
from telegrinder.modules import logger
from telegrinder.rules import Text

api = API(token=Token.from_env())
bot = Telegrinder(api)
logger.set_level("INFO")


@bot.on.message(Text("/ping"))
async def pong(m: Message):
    await m.answer("Pong")


bot.run_forever()



from __future__ import annotations

import typing
from abc import ABC, abstractmethod

if typing.TYPE_CHECKING:
    from telegrinder.api.api import API
    from telegrinder.bot.cute_types.base import BaseCute
    from telegrinder.bot.dispatch.waiter_machine.hasher.hasher import Hasher


class ABCScenario[Event: BaseCute](ABC):
    @abstractmethod
    def wait(self, hasher: Hasher, api: API) -> typing.Any:
        pass


__all__ = ("ABCScenario",)



import pytest

from telegrinder.bot.cute_types import CallbackQueryCute, MessageCute
from telegrinder.bot.dispatch.context import Context
from telegrinder.rules import ABCRule, IsPrivate, PayloadEqRule, check_rule
from telegrinder.types.objects import CallbackQuery, Message

message_event_with_text = {
    "message_id": 9999,
    "from": {
        "id": 1,
        "is_bot": False,
        "first_name": "Alex",
        "last_name": "Doe",
        "username": "alex777",
        "language_code": "en",
    },
    "chat": {
        "id": 1,
        "first_name": "Alex",
        "last_name": "Doe",
        "is_bot": False,
        "username": "alex777",
        "type": "private",
    },
    "date": 1234567898,
    "text": "",
}

cb_event_with_data = {
    "id": "4382bfdwdsb323b2d9",
    "from": {
        "id": 1,
        "is_bot": False,
        "first_name": "Alex",
        "last_name": "Doe",
        "username": "alex777",
        "language_code": "en",
    },
    "message": message_event_with_text,
    "chat_instance": "23asinstance23442",
    "inline_message_id": "1234csdbsk4839",
    "data": "",
}


class Text(ABCRule):
    def __init__(self, text: str) -> None:
        self.text = text

    def check(self, message: MessageCute) -> bool:
        return message.text.unwrap_or_none() == self.text


def test_rule_callback_data_eq(api_instance):
    cb_event_with_data["data"] = "test"
    cb_update = CallbackQuery.from_dict(cb_event_with_data)
    cb_event = CallbackQueryCute.from_update(cb_update, api_instance)
    assert PayloadEqRule("test").check(cb_event.data.unwrap()) is True
    assert PayloadEqRule("test1").check(cb_event.data.unwrap()) is False


def test_rule_text(api_instance):
    message_event_with_text["text"] = "hello!!!"
    msg_update = Message.from_dict(message_event_with_text)
    msg_event = MessageCute.from_update(msg_update, api_instance)
    assert Text("hello!!!").check(msg_event) is True
    assert Text("hello!!").check(msg_event) is False


@pytest.mark.asyncio()
async def test_rule_is_private_message_source(api_instance, message_update):
    assert await check_rule(api_instance, IsPrivate(), message_update, Context())


@pytest.mark.asyncio()
async def test_rule_is_private_callback_query_source(api_instance, callback_query_update):
    assert await check_rule(api_instance, IsPrivate(), callback_query_update, Context())



from telegrinder.bot.cute_types.pre_checkout_query import PreCheckoutQueryCute
from telegrinder.bot.rules.abc import ABCRule
from telegrinder.types.enums import Currency


class PaymentInvoiceCurrency(ABCRule):
    def __init__(self, currency: Currency, /) -> None:
        self.currency = currency

    def check(self, query: PreCheckoutQueryCute) -> bool:
        return self.currency == query.currency


__all__ = ("PaymentInvoiceCurrency",)



from __future__ import annotations

import typing

from fntypes.result import Error, Ok

from telegrinder.api.api import API
from telegrinder.bot.dispatch.context import Context
from telegrinder.bot.dispatch.middleware.abc import ABCMiddleware, run_middleware
from telegrinder.modules import logger
from telegrinder.node.composer import CONTEXT_STORE_NODES_KEY, compose_nodes
from telegrinder.tools.aio import maybe_awaitable
from telegrinder.tools.fullname import fullname
from telegrinder.tools.magic.function import bundle
from telegrinder.types.objects import Update

if typing.TYPE_CHECKING:
    from telegrinder.bot.dispatch.view.base import BaseView
    from telegrinder.bot.rules.abc import ABCRule


async def process_inner(
    api: API,
    event: Update,
    ctx: Context,
    view: BaseView,
) -> bool:
    ctx[CONTEXT_STORE_NODES_KEY] = {}  # For per-event shared nodes

    for m in view.middlewares:
        if (
            type(m).pre is not ABCMiddleware.pre
            and await run_middleware(m.pre, api, event, ctx, required_nodes=m.pre_required_nodes) is False
        ):
            logger.info(
                "Update(id={}, type={!r}) processed with view `{}`. Pre-middleware `{}` raised failure.",
                event.update_id,
                event.update_type,
                type(view).__name__,
                fullname(m),
            )
            return False

    found_handlers = []
    responses = list[typing.Any]()
    ctx_copy = ctx.copy()

    for handler in view.handlers:
        match await handler.run(api, event, ctx):
            case Ok(response):
                found_handlers.append(handler)
                responses.append(response)

                if view.return_manager is not None:
                    await view.return_manager.run(response, api, event, ctx)

                if handler.final is True:
                    break
            case Error(error):
                logger.debug(error)

    ctx = ctx_copy
    ctx.responses = responses

    for m in view.middlewares:
        if type(m).post is not ABCMiddleware.post:
            await run_middleware(
                m.post,
                api,
                event,
                ctx,
                required_nodes=m.post_required_nodes,
            )

    logger.info(
        "Update(id={}, type={!r}) processed with view `{}`. {}",
        event.update_id,
        event.update_type,
        type(view).__name__,
        "No found corresponded handlers."
        if not found_handlers
        else f"Handler{'s' if len(found_handlers) > 1 else ''}: {', '.join(f'`{x!r}`' for x in found_handlers)}",
    )
    return bool(found_handlers)


async def check_rule(
    api: API,
    rule: ABCRule,
    update: Update,
    ctx: Context,
) -> bool:
    """Checks requirements, adapts update.
    Returns check result.
    """
    # Running subrules to fetch requirements
    ctx_copy = ctx.copy()
    for requirement in rule.requires:
        if not await check_rule(api, requirement, update, ctx_copy):
            return False

    ctx |= ctx_copy
    node_col = None
    data = {Update: update, API: api, Context: ctx}

    # Composing required nodes
    if rule.required_nodes:
        match await compose_nodes(rule.required_nodes, ctx, data=data):
            case Ok(value):
                node_col = value
            case Error(compose_error):
                logger.debug(
                    "Cannot compose nodes for rule `{!r}`, error: {!r}",
                    rule,
                    compose_error.message,
                )
                return False

    # Running check
    try:
        bundle_check = bundle(rule.check, data, typebundle=True)
        bundle_check &= bundle(rule.check, ctx | ({} if node_col is None else node_col.values))
        return await maybe_awaitable(bundle_check())
    finally:
        # Closing node sessions if there are any
        if node_col is not None:
            await node_col.close_all()


__all__ = ("check_rule", "process_inner")



import asyncio
from inspect import isasyncgen, isawaitable

import typing_extensions as typing

type Generator[Yield, Send, Return] = typing.AsyncGenerator[Yield, Send] | typing.Generator[Yield, Send, Return]


def run_task[T](
    task: typing.Awaitable[T],
    /,
    *,
    loop: asyncio.AbstractEventLoop | None = None,
) -> T:
    loop = loop or asyncio.get_event_loop()
    return loop.run_until_complete(future=task)


async def next_generator[T](generator: Generator[T, typing.Any, typing.Any], /) -> T:
    return await send_generator_value(generator, None)


async def send_generator_value[Yield, Send](
    generator: Generator[Yield, Send, typing.Any],
    value: Send | None,
    /,
) -> Yield:
    try:
        return (
            await generator.asend(value) if isasyncgen(generator) else generator.send(value)  # type: ignore
        )
    except StopIteration as exc:
        raise StopGenerator(exc.value, exc.args) from exc


async def stop_generator[Send, Return](
    generator: Generator[typing.Any, Send, Return],
    with_value: Send | None = None,
    /,
) -> Return | None:
    try:
        await send_generator_value(generator, with_value)
    except (StopGenerator, StopAsyncIteration) as exc:
        return exc.value if isinstance(exc, StopGenerator) else None


async def maybe_awaitable[T](obj: T | typing.Awaitable[T], /) -> T:
    if isawaitable(obj):
        return await obj
    return obj


# Source code: https://github.com/facebookincubator/later/blob/main/later/task.py#L75
async def cancel_future(fut: asyncio.Future[typing.Any], /) -> None:
    if fut.done():
        return

    fut.cancel()
    exc: asyncio.CancelledError | None = None

    while not fut.done():
        shielded = asyncio.shield(fut)
        try:
            await asyncio.wait([shielded])
        except asyncio.CancelledError as ex:
            exc = ex
        finally:
            # Insure we handle the exception/value that may exist on the shielded task
            # This will prevent errors logged to the asyncio logger
            if shielded.done() and not shielded.cancelled() and not shielded.exception():
                shielded.result()

    if fut.cancelled():
        if exc is None:
            return
        raise exc from None

    ex = fut.exception()
    if ex is not None:
        raise ex from None

    raise asyncio.InvalidStateError(
        f"Task did not raise CancelledError on cancel: {fut!r} had result {fut.result()!r}",
    )


class StopGenerator(Exception):
    value: typing.Any

    def __init__(self, value: typing.Any, *args: object) -> None:
        super().__init__(*args)
        self.value = value


__all__ = (
    "StopGenerator",
    "cancel_future",
    "maybe_awaitable",
    "next_generator",
    "run_task",
    "send_generator_value",
    "stop_generator",
)



from fntypes.result import RESULT_ERROR_LOGGER, Error

from telegrinder import API, Message, Telegrinder, Token
from telegrinder.modules import logger
from telegrinder.rules import Text

logger.set_level("INFO")
RESULT_ERROR_LOGGER.set_log(logger.error)

api = API(token=Token.from_env())
bot = Telegrinder(api)


@bot.on.message(Text("/error_no_control"))
async def handle_error_no_control(m: Message):
    # This is going to log potentially lost information about api error
    # Here error is not considered to be processed
    await api.send_message(chat_id=m.chat.id, text="")

    # In the following context error is also considered not to be processed
    # because .unwrap_or method omits the error data in the case of Error state of result
    # Thus, nothing is going to be logged
    result = await api.send_message(chat_id=m.chat.id, text="")
    result.unwrap_or(None)


@bot.on.message(Text("/error_controlled"))
async def handle_error_controlled(m: Message):
    # This is not going to log api error traceback because error is being processed with .unwrap()
    # Here error is considered to be processed in the scope we are in
    (await api.send_message(chat_id=m.chat.id, text="")).unwrap()

    # This is going to raise exception
    print("Unreachable code")


@bot.on.message(Text("/error_manually_controlled"))
async def handle_error_manually_controlled(m: Message):
    # This is not going to log api error due to the access (two cases demonstrated) of error field
    # Here error is considered to be processed
    result = await api.send_message(chat_id=m.chat.id, text="")

    # First case, control pass with pattern matching
    match result:
        case Error(err):
            print(err)

    # Second case
    if isinstance(result, Error):
        print(result.error)

    # Thus, no log is going to happen, only two prints out of cases shown above


bot.run_forever()



import typing


def join_dicts[Key: typing.Hashable, Value](
    left_dict: dict[Key, typing.Any],
    right_dict: dict[typing.Any, Value],
    /,
) -> dict[Key, Value]:
    return {left_key: right_dict[right_key] for left_key, right_key in left_dict.items()}


def extract[Key: typing.Hashable, Value](
    keys: typing.Iterable[Key],
    mapping: typing.Mapping[Key, Value],
    /,
) -> dict[Key, Value]:
    return {key: mapping[key] for key in keys if key in mapping}


__all__ = ("extract", "join_dicts")
